/*
    CharityWorld: 3x3charity.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*

3x3charity.m

This is a cut-down version of charityWorld with no display, a 3x3 grid,
SSAs, Moore neighbourhood radius 1, the winner in the centre of the grid,
using epsilons, and generosity = 0.4, initial wealth 0.0.

*/

#import <stdio.h>
#import <stdlib.h>
#import <string.h>
#import <float.h>
#ifndef __CYGWIN__
#  import <ieeefp.h>
#else
#  ifdef INTEL
#    import <CIieeefp.h>
#  else
#    error "This program requires IEEE floating point arithmetic support"
#  endif
#endif
#import <math.h>
#import <objc/objc.h>

#import "../DoubleEpsilon.h"
#import "../DoubleIntervalWarn.h"
#import "../DoubleKahanSum.h"
#import "../DoubleLong.h"
#import "../DoubleOffset.h"
#import "../DoubleShort.h"
#import "../DoubleSimple.h"
#import "../DoubleString.h"
#import "../DoubleWarn.h"

#define NAGENTS 9

typedef enum { UNDEFINED, POINT_FOUR, POINT_FIVE } GENEROSITY_OPTION;

#define GENEROSITY_STR_4 "0.4"
#define CALC_TRANSFERS_4(j) (20.0 * j) / 9.0

#define GENEROSITY_STR_5 "0.5"
#define CALC_TRANSFERS_5(j) (16.0 * j) / 9.0

  /* transfers = ((NAGENTS - 1) * jackpot * GEN_DENOM) / (GEN_NUM * NAGENTS) */

typedef struct {
  DoubleEpsilon *epsilon;
  DoubleIntervalWarn *interval;
  DoubleKahanSum *kahan;
  DoubleLong *extend;
  DoubleOffset *offset;
  DoubleOffset *offoffset;
  DoubleShort *single;
  DoubleSimple *simple;
  DoubleString *string;
  DoubleWarn *warn;
} DOUBLE;

typedef struct {
  BOOL epsilon;
  BOOL interval;
  BOOL kahan;
  BOOL extend;
  BOOL offset;
  BOOL offoffset;
  BOOL single;
  BOOL simple;
  BOOL string;
  BOOL warn;
} BOOL_DOUBLE;

const BOOL_DOUBLE YES_DOUBLE = {
  YES, YES, YES, YES, YES, YES, YES, YES, YES, YES
};

//eps, ivl, khn, ext, off, ow0, sgl, sim, str, wrn

const BOOL_DOUBLE NO_DOUBLE = {
   NO,  NO,  NO,  NO,  NO,  NO,  NO,  NO,  NO,  NO
};

#define DO_ME(s, me, msg, arg) \
  if(s->epsilon) [me->epsilon msg: arg]; \
  if(s->interval) [me->interval msg: arg]; \
  if(s->kahan) [me->kahan msg: arg]; \
  if(s->extend) [me->extend msg: arg]; \
  if(s->offset) [me->offset msg: arg]; \
  if(s->offoffset) [me->offoffset msg: arg]; \
  if(s->single) [me->single msg: arg]; \
  if(s->simple) [me->simple msg: arg]; \
  if(s->string) [me->string msg: arg]; \
  if(s->warn) [me->warn msg: arg]

#define DO_ME_NO_ARG(s, me, msg) \
  if(s->epsilon) [me->epsilon msg]; \
  if(s->interval) [me->interval msg]; \
  if(s->kahan) [me->kahan msg]; \
  if(s->extend) [me->extend msg]; \
  if(s->offset) [me->offset msg]; \
  if(s->offoffset) [me->offoffset msg]; \
  if(s->single) [me->single msg]; \
  if(s->simple) [me->simple msg]; \
  if(s->string) [me->string msg]; \
  if(s->warn) [me->warn msg]

#define DO_ME_CONST(ss, me, msg, arg) \
  if(ss->epsilon) [me->epsilon msg: (double)arg]; \
  if(ss->interval) [me->interval msg: (double)arg]; \
  if(ss->kahan) [me->kahan msg: (double)arg]; \
  if(ss->extend) [me->extend l ## msg: arg]; \
  if(ss->offset) [me->offset msg: (double)arg]; \
  if(ss->offoffset) [me->offoffset msg: (double)arg]; \
  if(ss->single) [me->single s ## msg: (float)arg]; \
  if(ss->simple) [me->simple msg: (double)arg]; \
  if(ss->string) [me->string msg: (double)arg]; \
  if(ss->warn) [me->warn msg: (double)arg]

#define DO_ME_ARG(s, me, msg, arg) \
  if(s->epsilon) [me->epsilon msg: arg->epsilon]; \
  if(s->interval) [me->interval msg: arg->interval]; \
  if(s->kahan) [me->kahan msg: arg->kahan]; \
  if(s->extend) [me->extend msg: arg->extend]; \
  if(s->offset) [me->offset msg: arg->offset]; \
  if(s->offoffset) [me->offoffset msg: arg->offoffset]; \
  if(s->single) [me->single msg: arg->single]; \
  if(s->simple) [me->simple msg: arg->simple]; \
  if(s->string) [me->string msg: arg->string]; \
  if(s->warn) [me->warn msg: arg->warn]

#define Q_DO_ME_ARG(s, q, me, msg, arg) \
  if(s->epsilon) if(q->epsilon) [me->epsilon msg: arg->epsilon]; \
  if(s->interval) if(q->interval) [me->interval msg: arg->interval]; \
  if(s->kahan) if(q->kahan) [me->kahan msg: arg->kahan]; \
  if(s->extend) if(q->extend) [me->extend msg: arg->extend]; \
  if(s->offset) if(q->offset) [me->offset msg: arg->offset]; \
  if(s->offoffset) if(q->offoffset) [me->offoffset msg: arg->offoffset]; \
  if(s->single) if(q->single) [me->single msg: arg->single]; \
  if(s->simple) if(q->simple) [me->simple msg: arg->simple]; \
  if(s->string) if(q->string) [me->string msg: arg->string]; \
  if(s->warn) if(q->warn) [me->warn msg: arg->warn]

#define DO_ME_CLASS_NO_ARG(s, me, msg) \
  if(s->epsilon) me->epsilon = [DoubleEpsilon msg]; \
  if(s->interval) me->interval = [DoubleIntervalWarn msg]; \
  if(s->kahan) me->kahan = [DoubleKahanSum msg]; \
  if(s->extend) me->extend = [DoubleLong msg]; \
  if(s->offset) me->offset = [DoubleOffset msg]; \
  if(s->offoffset) me->offoffset = [DoubleOffset msg]; \
  if(s->single) me->single = [DoubleShort msg]; \
  if(s->simple) me->simple = [DoubleSimple msg]; \
  if(s->string) me->string = [DoubleString msg]; \
  if(s->warn) me->warn = [DoubleWarn msg]

#define TEST_ME_ARG(s, ans, me, msg, arg) \
  if(s->epsilon) ans->epsilon = [me->epsilon msg: arg->epsilon]; \
  if(s->interval) ans->interval = [me->interval msg: arg->interval]; \
  if(s->kahan) ans->kahan = [me->kahan msg: arg->kahan]; \
  if(s->extend) ans->extend = [me->extend msg: arg->extend]; \
  if(s->offset) ans->offset = [me->offset msg: arg->offset]; \
  if(s->offoffset) ans->offoffset = [me->offoffset msg: arg->offoffset]; \
  if(s->single) ans->single = [me->single msg: arg->single]; \
  if(s->simple) ans->simple = [me->simple msg: arg->simple]; \
  if(s->string) ans->string = [me->string msg: arg->string]; \
  if(s->warn) ans->warn = [me->warn msg: arg->warn]

#define Q_PRINT_MSG(s, q, msg) \
  if(s->epsilon) if(q->epsilon) printf("DoubleEpsilon: %s\n", msg); \
  if(s->interval) if(q->interval) printf("DoubleIntervalWarn: %s\n", msg); \
  if(s->kahan) if(q->kahan) printf("DoubleKahanSum: %s\n", msg); \
  if(s->extend) if(q->extend) printf("DoubleLong: %s\n", msg); \
  if(s->offset) if(q->offset) printf("DoubleOffset (w0 == 0): %s\n", msg); \
  if(s->offoffset) \
    if(q->offoffset) printf("DoubleOffset (w0 > 0): %s\n", msg); \
  if(s->single) if(q->single) printf("DoubleShort: %s\n", msg); \
  if(s->simple) if(q->simple) printf("DoubleSimple: %s\n", msg); \
  if(s->string) if(q->string) printf("DoubleString: %s\n", msg); \
  if(s->warn) if(q->warn) printf("DoubleWarn: %s\n", msg)

#define PRINT_VAL_MSG(s, val, msg) \
  printf(msg); \
  if(s->epsilon) { \
    printf("\n\tDoubleEpsilon: "); \
    [val->epsilon writeAccurately: stdout]; \
  } \
  if(s->interval) { \
    printf("\n\tDoubleIntervalWarn: "); \
    [val->interval writeAccurately: stdout]; \
  } \
  if(s->kahan) { \
    printf("\n\tDoubleKahanSum: "); \
    [val->kahan writeAccurately: stdout]; \
  } \
  if(s->extend) { \
    printf("\n\tDoubleLong: "); \
    [val->extend writeAccurately: stdout]; \
  } \
  if(s->offset) { \
    printf("\n\tDoubleOffset (w0 == 0): "); \
    [val->offset writeAccurately: stdout]; \
  } \
  if(s->offoffset) { \
    printf("\n\tDoubleOffset (w0 > 0): "); \
    [val->offoffset writeAccurately: stdout]; \
  } \
  if(s->single) { \
    printf("\n\tDoubleShort: "); \
    [val->single writeAccurately: stdout]; \
  } \
  if(s->simple) { \
    printf("\n\tDoubleSimple: "); \
    [val->simple writeAccurately: stdout]; \
  } \
  if(s->string) { \
    printf("\n\tDoubleString: "); \
    [val->string writeAccurately: stdout]; \
  } \
  if(s->warn) { \
    printf("\n\tDoubleWarn: "); \
    [val->warn writeAccurately: stdout]; \
  } \
  printf("\n")

void and(BOOL_DOUBLE *some, BOOL_DOUBLE *ans, BOOL_DOUBLE *arg1,
	 BOOL_DOUBLE *arg2) {
  if(some->epsilon) ans->epsilon = (arg1->epsilon && arg2->epsilon ? YES : NO);
  if(some->interval)
    ans->interval = (arg1->interval && arg2->interval ? YES : NO);
  if(some->kahan) ans->kahan = (arg1->kahan && arg2->kahan ? YES : NO);
  if(some->extend) ans->extend = (arg1->extend && arg2->extend ? YES : NO);
  if(some->offset) ans->offset = (arg1->offset && arg2->offset ? YES : NO);
  if(some->offoffset)
    ans->offoffset = (arg1->offoffset && arg2->offoffset ? YES : NO);
  if(some->single) ans->single = (arg1->single && arg2->single ? YES : NO);
  if(some->simple) ans->simple = (arg1->simple && arg2->simple ? YES : NO);
  if(some->string) ans->string = (arg1->string && arg2->string ? YES : NO);
  if(some->warn) ans->warn = (arg1->warn && arg2->warn ? YES : NO);
}

void or(BOOL_DOUBLE *some,
	BOOL_DOUBLE *ans, BOOL_DOUBLE *arg1, BOOL_DOUBLE *arg2) {
  if(some->epsilon) ans->epsilon = (arg1->epsilon || arg2->epsilon ? YES : NO);
  if(some->interval)
    ans->interval = (arg1->interval || arg2->interval ? YES : NO);
  if(some->kahan) ans->kahan = (arg1->kahan || arg2->kahan ? YES : NO);
  if(some->extend) ans->extend = (arg1->extend || arg2->extend ? YES : NO);
  if(some->offset) ans->offset = (arg1->offset || arg2->offset ? YES : NO);
  if(some->offoffset)
    ans->offoffset = (arg1->offoffset || arg2->offoffset ? YES : NO);
  if(some->single) ans->single = (arg1->single || arg2->single ? YES : NO);
  if(some->simple) ans->simple = (arg1->simple || arg2->simple ? YES : NO);
  if(some->string) ans->string = (arg1->string || arg2->string ? YES : NO);
  if(some->warn) ans->warn = (arg1->warn || arg2->warn ? YES : NO);
}

void not(BOOL_DOUBLE *some, BOOL_DOUBLE *ans, BOOL_DOUBLE *arg) {
  if(some->epsilon) ans->epsilon = (arg->epsilon ? NO : YES);
  if(some->interval)
    ans->interval = (arg->interval ? NO : YES);
  if(some->kahan) ans->kahan = (arg->kahan ? NO : YES);
  if(some->extend) ans->extend = (arg->extend ? NO : YES);
  if(some->offset) ans->offset = (arg->offset ? NO : YES);
  if(some->offoffset)
    ans->offoffset = (arg->offoffset ? NO : YES);
  if(some->single) ans->single = (arg->single ? NO : YES);
  if(some->simple) ans->simple = (arg->simple ? NO : YES);
  if(some->string) ans->string = (arg->string ? NO : YES);
  if(some->warn) ans->warn = (arg->warn ? NO : YES);
}

void check_warn(BOOL_DOUBLE *some, BOOL_DOUBLE *ans, DOUBLE *number) {
  (*ans) = NO_DOUBLE;
  if(some->warn) ans->warn = [number->warn detectedAny];
  if(some->interval) ans->interval = [number->interval detectedError];
}

void init(BOOL_DOUBLE *some, DOUBLE *me, char *value) {
  DO_ME_CLASS_NO_ARG(some, me, new);
  DO_ME(some, me, setStr, value);
}

void kill(BOOL_DOUBLE *some, DOUBLE *me) {
  DO_ME_NO_ARG(some, me, free);
}

void setStr(BOOL_DOUBLE *some, DOUBLE *me, char *value) {
  DO_ME(some, me, setStr, value);
}

void set(BOOL_DOUBLE *some, DOUBLE *me, DOUBLE *value) {
  DO_ME_ARG(some, me, set, value);
}

void qprint(BOOL_DOUBLE *some, BOOL_DOUBLE *q, char *msg) {
  Q_PRINT_MSG(some, q, msg);
}

void qprint_cycle(BOOL_DOUBLE *some, BOOL_DOUBLE *q, char *msg, double cycle) {
  char *truemsg, tempbuf[10];	
  int size = snprintf(tempbuf, 9, "%s (cycle %.*e)", msg, DBL_DIG - 1, cycle);

  truemsg = (char *)malloc((size + 1) * sizeof(char));
  if(truemsg == NULL) {
    perror("Memory allocation of truemsg");
    abort();
  }
  sprintf(truemsg, "%s (cycle %.*e)", msg, DBL_DIG - 1, cycle);
  Q_PRINT_MSG(some, q, truemsg);
  free(truemsg);
}

void print_values(BOOL_DOUBLE *some, DOUBLE *value, const char *msg) {
  PRINT_VAL_MSG(some, value, msg);
}

void add(BOOL_DOUBLE *some, DOUBLE *me, DOUBLE *value) {
  DO_ME_ARG(some, me, add, value);
}

void sub(BOOL_DOUBLE *some, DOUBLE *me, DOUBLE *value) {
  DO_ME_ARG(some, me, sub, value);
}

void qadd(BOOL_DOUBLE *some, BOOL_DOUBLE *q, DOUBLE *me, DOUBLE *value) {
  Q_DO_ME_ARG(some, q, me, add, value);
}

void qsub(BOOL_DOUBLE *some, BOOL_DOUBLE *q, DOUBLE *me, DOUBLE *value) {
  Q_DO_ME_ARG(some, q, me, sub, value);
}

void cmul(BOOL_DOUBLE *some, DOUBLE *me, long double value) {
  DO_ME_CONST(some, me, cmul, value);
}

void cdiv(BOOL_DOUBLE *some, DOUBLE *me, long double value) {
  DO_ME_CONST(some, me, cdiv, value);
}

void gt(BOOL_DOUBLE *some, BOOL_DOUBLE *ans, DOUBLE *me, DOUBLE *value) {
  TEST_ME_ARG(some, ans, me, gt, value);
}

void lt(BOOL_DOUBLE *some, BOOL_DOUBLE *ans, DOUBLE *me, DOUBLE *value) {
  TEST_ME_ARG(some, ans, me, lt, value);
}

void get_selection(BOOL_DOUBLE *some, char *list) {
  if(strcmp(list, "all") == 0) {
    (*some) = YES_DOUBLE;
  }
  else {
    char *p, *q;

    (*some) = NO_DOUBLE;
    q = list;
    while((p = strchr(q, (int)',')) != NULL) {
      if(strncmp(q, "eps", p - q) == 0) some->epsilon = YES;
      else if(strncmp(q, "ivl", p - q) == 0) some->interval = YES;
      else if(strncmp(q, "khn", p - q) == 0) some->kahan = YES;
      else if(strncmp(q, "ext", p - q) == 0) some->extend = YES;
      else if(strncmp(q, "off", p - q) == 0) some->offset = YES;
      else if(strncmp(q, "ow0", p - q) == 0) some->offoffset = YES;
      else if(strncmp(q, "sgl", p - q) == 0) some->single = YES;
      else if(strncmp(q, "sim", p - q) == 0) some->simple = YES;
      else if(strncmp(q, "str", p - q) == 0) some->string = YES;
      else if(strncmp(q, "wrn", p - q) == 0) some->warn = YES;
      else {
	fprintf(stderr, "Invalid element in classlist at %s\n", q);
	abort();
      }
      q = p + 1;
    }
    if(strcmp(q, "eps") == 0) some->epsilon = YES;
    else if(strcmp(q, "ivl") == 0) some->interval = YES;
    else if(strcmp(q, "khn") == 0) some->kahan = YES;
    else if(strcmp(q, "ext") == 0) some->extend = YES;
    else if(strcmp(q, "off") == 0) some->offset = YES;
    else if(strcmp(q, "ow0") == 0) some->offoffset = YES;
    else if(strcmp(q, "sgl") == 0) some->single = YES;
    else if(strcmp(q, "sim") == 0) some->simple = YES;
    else if(strcmp(q, "str") == 0) some->string = YES;
    else if(strcmp(q, "wrn") == 0) some->warn = YES;
    else {
      fprintf(stderr, "Invalid element in classlist at %s\n", q);
      abort();
    }
  }
}
   
int getmember(char *txt, const char *arr[]) {
  int i;

  for(i = 0; arr[i][0] != '\0'; i++) {
    if(strcmp(txt, arr[i]) == 0) return i;
  }
  return -1;
}

int main(int argc, char **argv) {
  DOUBLE winner_wealth;
  DOUBLE losers_wealth;
  DOUBLE losers_wealth_arr[NAGENTS - 1];
  /* Loser arrangement:
   *
   *			0	1	2
   *
   *			3	Winner	4
   *
   *			5	6	7
   */
  int loser_planarnbr_arr[NAGENTS - 1][5] = { { 1,  3, -1, -1, -1 },
					      { 0,  3,  2,  4, -1 },
					      { 1,  4, -1, -1, -1 },
					      { 0,  1,  5,  6, -1 },
					      { 1,  2,  6,  7, -1 },
					      { 3,  6, -1, -1, -1 },
					      { 3,  4,  5,  7, -1 },
					      { 4,  6, -1, -1, -1 } };
  int loser_toroidalnbr_arr[NAGENTS - 1][NAGENTS - 1] = {
    { 1, 2, 3, 4, 5, 6, 7, -1 },
    { 0, 2, 3, 4, 5, 6, 7, -1 },
    { 0, 1, 3, 4, 5, 6, 7, -1 },
    { 0, 1, 2, 4, 5, 6, 7, -1 },
    { 0, 1, 2, 3, 5, 6, 7, -1 },
    { 0, 1, 2, 3, 4, 6, 7, -1 },
    { 0, 1, 2, 3, 4, 5, 7, -1 },
    { 0, 1, 2, 3, 4, 5, 6, -1 }
  };
  BOOL toroidal = YES;
  DOUBLE jackpot;
  DOUBLE ticket_price;
  DOUBLE total_wealth;
  DOUBLE expr_temp;
  double temp;
  int a, i;
  DOUBLE generosity;
  double transfers;
  double onepctxfer, npctxfer;
  double offset_init_wealth;
  BOOL_DOUBLE awareness, approach, exchange, bool_temp, selection;
  BOOL_DOUBLE warnings, validity, this_valid, this_warn;
  GENEROSITY_OPTION gen_choice = UNDEFINED;
  BOOL show_wealth = NO;

  if(argc < 8) {
    fprintf(stderr, "Usage: %s <classlist> <0.4|0.5> <jackpot> <epsilon> "
	    "<offset> <sigfig> <toroidal|planar> [showWealth]\n", argv[0]);
    fprintf(stderr, "\tClasslist may be one of:\n\t\tall\tAll classes\n"
	    "\t\teps\tDoubleEpsilon\n\t\tivl\tDoubleInterval\n"
	    "\t\tkhn\tDoubleKahan\n\t\text\tDoubleLong\n"
	    "\t\toff\tDoubleOffset (w0 = 0)\n\t\tow0\tDoubleOffset (w0 > 0)\n"
	    "\t\tsgl\tDoubleShort\n\t\tsim\tDoubleSimple\n"
	    "\t\tstr\tDoubleString\n\t\twrn\tDoubleWarn\n");
    fprintf(stderr, "\tElements in the classlist should be comma separated\n");
    fprintf(stderr, "\n\tepsilon should be n/a if DoubleEpsilon is not in "
	    "classlist\n\toffset should be n/a if DoubleOffset is not in "
	    "classlist\n\tsigfig should be n/a if DoubleString is not in "
	    "classlist\n");
    fprintf(stderr, "\nIf showWealth is the last argument, then the wealths of"
	    " all the agents will be\nshown at the end of the run\n");
    abort();
  }
  else {
    int i;
    BOOL more_argnames = YES;
    const char *argnames[] = {
      "Command", "Classlist", "Generosity", "Jackpot", "Epsilon", "Offset", 
      "Significant figures", "Topology", ""
    };
    const char *argopts[] = {
      "showWealth", ""
    };
    BOOL *optvars[] = { &show_wealth };

    for(i = 0; i < argc; i++) {
      const char *paramname;

      more_argnames = more_argnames
	? (argnames[i][0] == '\0' ? NO : YES)
	: more_argnames;
      if(more_argnames) {
	paramname = argnames[i];
      }
      else {
	int j = getmember(argv[i], argopts);

	if(j >= 0) {
	  paramname = "(option)";
	  *(optvars[j]) = YES;
	}
	else {
	  paramname = "(ignored)";
	}
      }

      printf("%s: %s\n", paramname, argv[i]);
    }
  }

  get_selection(&selection, argv[1]);

  if(strcmp(argv[2], "0.4") == 0) {
    gen_choice = POINT_FOUR;
  }
  else if(strcmp(argv[2], "0.5") == 0) {
    gen_choice = POINT_FIVE;
  }
  else {
    fprintf(stderr, "Invalid generosity choice: %s\n", argv[2]);
    abort();
  }

  if(strcmp(argv[7], "toroidal") == 0) {
    toroidal = YES;
  }
  else if(strcmp(argv[7], "planar") == 0) {
    toroidal = NO;
  }
  else {
    fprintf(stderr, "Invalid environment topology: %s\n", argv[7]);
    abort();
  }

  /* Initialise DOUBLES */

  if(selection.epsilon) {
    [DoubleEpsilon setEpsilon: atof(argv[4])];
  }
  else {
    if(strcmp(argv[4], "n/a") != 0) {
      printf("WARNING: Epsilon %s ignored\n", argv[4]);
    }
  }

  if(selection.offset || selection.offoffset) {
    [DoubleOffset setOffset: atof(argv[5])];
  }
  else {
    if(strcmp(argv[5], "n/a") != 0) {
      printf("WARNING: Offset %s ignored\n", argv[5]);
    }
  }

  if(selection.string) {
    [DoubleString setSignificantFigures: atof(argv[6])];
  }
  else {
    if(strcmp(argv[6], "n/a") != 0) {
      printf("WARNING: Significant figures %s ignored\n", argv[6]);
    }
  }

  [DoubleWarn trapNone];	/* Switch off printing a warning message with
				 * each fperr detected. Fperrs will still be
				 * recorded by DoubleWarn's superclass
				 * DoubleDetect.
				 */
  init(&selection, &jackpot, argv[3]);
  switch(gen_choice) {
  case POINT_FOUR:
    init(&selection, &generosity, GENEROSITY_STR_4);
    break;
  case POINT_FIVE:
    init(&selection, &generosity, GENEROSITY_STR_5);
    break;
  }
  init(&selection, &winner_wealth, "0.0");
  init(&selection, &losers_wealth, "0.0");
  init(&selection, &ticket_price, "0.0");
  init(&selection, &expr_temp, "0.0");

  warnings = NO_DOUBLE;
  validity = YES_DOUBLE;

  /* Initialise wealth distribution */

  set(&selection, &ticket_price, &jackpot);
  cdiv(&selection, &ticket_price, (long double)NAGENTS);
  check_warn(&selection, &bool_temp, &ticket_price);
  or(&selection, &warnings, &warnings, &bool_temp);

  if(selection.offoffset) {
    offset_init_wealth = ceil([ticket_price.offoffset get]);
    [winner_wealth.offoffset cset: offset_init_wealth];
    [losers_wealth.offoffset cset: offset_init_wealth];
  }
  
  sub(&selection, &losers_wealth, &ticket_price);
  check_warn(&selection, &bool_temp, &losers_wealth);
  or(&selection, &warnings, &warnings, &bool_temp);
  for(a = 1; a < NAGENTS; a++) {
    add(&selection, &winner_wealth, &ticket_price);
    check_warn(&selection, &bool_temp, &winner_wealth);
    or(&selection, &warnings, &warnings, &bool_temp);
  }
  
  //  not(&selection, &bool_temp, &warnings);
  //  and(&selection, &validity, &validity, &bool_temp);

  qprint(&selection, &warnings, "warning issued during initialisation");

  /* Calculate the correct number of transfers. We need to know that
     this is accurate. */

  temp = atof(argv[3]);
  switch(gen_choice) {
  case POINT_FOUR:
    transfers = rint(CALC_TRANSFERS_4(temp));
    break;
  case POINT_FIVE:
    transfers = rint(CALC_TRANSFERS_5(temp));
    break;
  }
  onepctxfer = transfers / 100.0;
  npctxfer = transfers;

  printf("The model should converge after %.*e transfers, or %.*e cycles\n",
	 DBL_DIG, transfers, DBL_DIG, transfers / (double)(NAGENTS - 1));

  /* Check that subtracting (NAGENTS - 1) from transfers can be
     achieved without loss of accuracy */

  fpsetsticky(0);
  temp = transfers - (double)(NAGENTS - 1);
  if(fpgetsticky() != 0) {
    fprintf(stderr, "The number of transfers is too large to subtract %d from"
	    " it accurately\n", NAGENTS - 1);
    abort();
  }

  for(i = 0; i < NAGENTS - 1; i++) {
    init(&selection, &losers_wealth_arr[i], "0.0");
    set(&selection, &losers_wealth_arr[i], &losers_wealth);
  }
  
  /* Redistribute the wealth */

  init(&selection, &total_wealth, "0.0");
  for(;
      transfers > 0.0;
      transfers -= (double)(NAGENTS - 1)) {

    if(transfers < npctxfer) {
      npctxfer -= onepctxfer;
      printf("Progress: %g%%\n", 100.0 - (npctxfer / onepctxfer));
      fflush(stdout);
    }

    /* print wealths if required */
    
    if(show_wealth && (transfers / (double)(NAGENTS - 1)) == 1.0) {
      printf("Wealth before last cycle is undertaken:\n");
      print_values(&selection, &winner_wealth, "Winner's wealth:");

      for(i = 0; i < NAGENTS - 1; i++) {
	printf("Loser %d's ", i);
	print_values(&selection, &losers_wealth_arr[i], "wealth:");
      }
    }

    /* continue redistributing wealth */

    for(i = 0; i < NAGENTS - 1; i++) {
      int j;

      this_valid = YES_DOUBLE;
      this_warn = NO_DOUBLE;

      set(&selection, &total_wealth, &winner_wealth);

      for(j = 0; j < NAGENTS - 1; j++) {
	add(&selection, &total_wealth, &losers_wealth_arr[j]);
	check_warn(&selection, &bool_temp, &total_wealth);
	or(&selection, &this_warn, &this_warn, &bool_temp);
      }

      /* test awareness condition */

      set(&selection, &expr_temp, &winner_wealth);
      cmul(&selection, &expr_temp, (long double)(NAGENTS));
      check_warn(&selection, &bool_temp, &expr_temp);
      or(&selection, &this_warn, &this_warn, &bool_temp);

      gt(&selection, &awareness, &expr_temp, &total_wealth);
      check_warn(&selection, &bool_temp, &expr_temp);
      or(&selection, &this_warn, &this_warn, &bool_temp);

      /* print a warning message if a new warning has been issued */

      not(&selection, &bool_temp, &warnings);
      and(&selection, &bool_temp, &bool_temp, &this_warn);
      qprint_cycle(&selection, &bool_temp,
		   "warning issued for awareness condition",
		   transfers / (double)(NAGENTS - 1));

      or(&selection, &warnings, &warnings, &this_warn);
				/* update warnings */

      //      not(&selection, &bool_temp, &warnings);
      //      and(&selection, &this_valid, &this_valid, &bool_temp);
      and(&selection, &this_valid, &this_valid, &awareness);
				/* update validity of this part of the cycle */

      /* print a warning message if the awareness condition is not satisfied */

      not(&selection, &bool_temp, &this_valid);
      and(&selection, &bool_temp, &bool_temp, &validity);
      qprint_cycle(&selection, &bool_temp, "awareness condition not satisfied",
		   transfers / (double)(NAGENTS - 1));

      and(&selection, &validity, &validity, &this_valid);
				/* update validity of the run */

      /* test approach condition */

      this_valid = YES_DOUBLE;
      this_warn = NO_DOUBLE;

      set(&selection, &expr_temp, &losers_wealth_arr[i]);
      cmul(&selection, &expr_temp, (long double)(NAGENTS));
      check_warn(&selection, &bool_temp, &expr_temp);
      or(&selection, &this_warn, &this_warn, &bool_temp);

      lt(&selection, &approach, &expr_temp, &total_wealth);
      check_warn(&selection, &bool_temp, &expr_temp);
      or(&selection, &this_warn, &this_warn, &bool_temp);

      not(&selection, &bool_temp, &warnings);
      and(&selection, &bool_temp, &bool_temp, &this_warn);
      qprint_cycle(&selection, &bool_temp,
		   "warning issued for approach condition",
		   transfers / (double)(NAGENTS - 1));

      or(&selection, &warnings, &warnings, &this_warn);

      //      not(&selection, &bool_temp, &warnings);
      //      and(&selection, &this_valid, &this_valid, &bool_temp);
      and(&selection, &this_valid, &this_valid, &approach);

      not(&selection, &bool_temp, &this_valid);
      and(&selection, &bool_temp, &bool_temp, &validity);
      qprint_cycle(&selection, &bool_temp, "approach condition not satisfied",
		   transfers / (double)(NAGENTS - 1));

      and(&selection, &validity, &validity, &this_valid);


      /* exchange the wealth */

      this_valid = YES_DOUBLE;
      this_warn = NO_DOUBLE;

      and(&selection, &exchange, &awareness, &approach);

      qsub(&selection, &exchange, &winner_wealth, &generosity);
      check_warn(&selection, &bool_temp, &winner_wealth);
      or(&selection, &this_warn, &this_warn, &bool_temp);

      qadd(&selection, &exchange, &losers_wealth_arr[i], &generosity);
      check_warn(&selection, &bool_temp, &losers_wealth_arr[i]);
      or(&selection, &this_warn, &this_warn, &bool_temp);

      not(&selection, &bool_temp, &warnings);
      and(&selection, &bool_temp, &bool_temp, &this_warn);
      qprint_cycle(&selection, &bool_temp,
		   "warning issued for wealth exchange",
		   transfers / (double)(NAGENTS - 1));

      //      not(&selection, &bool_temp, &warnings);
      //      and(&selection, &validity, &validity, &bool_temp);

    }
  }

  npctxfer -= onepctxfer;
  printf("Progress: %g%%\n", 100.0 - (npctxfer / onepctxfer));
  fflush(stdout);

  /* test awareness condition of all agents at the end */

  this_valid = YES_DOUBLE;
  this_warn = NO_DOUBLE;

  set(&selection, &total_wealth, &winner_wealth);
  for(i = 0; i < NAGENTS - 1; i++) {
    add(&selection, &total_wealth, &losers_wealth_arr[i]);	
    check_warn(&selection, &bool_temp, &total_wealth);
    or(&selection, &this_warn, &this_warn, &bool_temp);
  }

  /* test winning agent awareness condition */

  set(&selection, &expr_temp, &winner_wealth);
  cmul(&selection, &expr_temp, (long double)(NAGENTS));
  check_warn(&selection, &bool_temp, &expr_temp);
  or(&selection, &this_warn, &this_warn, &bool_temp);

  gt(&selection, &awareness, &expr_temp, &total_wealth);
  check_warn(&selection, &bool_temp, &expr_temp);
  or(&selection, &this_warn, &this_warn, &bool_temp);

  /* print a warning message if a new warning has been issued */

  not(&selection, &bool_temp, &warnings);
  and(&selection, &bool_temp, &bool_temp, &this_warn);
  qprint(&selection, &bool_temp,
	 "warning issued for winner awareness condition (after last cycle)");

  or(&selection, &warnings, &warnings, &this_warn);
				/* update warnings.
				 *
				 * IMPORTANT NOTE: Warnings after the last
				 * cycle do not impact the validity of the
				 * run. The warning messages are only
				 * printed for information.
				 */

  not(&selection, &bool_temp, &awareness);
  and(&selection, &this_valid, &this_valid, &bool_temp);
				/* update validity of this part of the cycle */

  /* print a warning message if the awareness condition is satisfied */

  not(&selection, &bool_temp, &this_valid);
  and(&selection, &bool_temp, &bool_temp, &validity);
  qprint(&selection, &bool_temp,
	 "winner awareness condition satisfied (after last cycle)");

  and(&selection, &validity, &validity, &this_valid);
				/* update validity of the run */

  /* test loser agents' awareness conditions */

  for(i = 0; i < NAGENTS - 1; i++) {
    DOUBLE loser_total;
    int loser_nagents, j;
    this_warn = NO_DOUBLE;
    this_valid = YES_DOUBLE;

    printf("%d: Winner", i);

    init(&selection, &loser_total, "0.0");
    set(&selection, &loser_total, &losers_wealth_arr[i]);
    add(&selection, &loser_total, &winner_wealth);
    check_warn(&selection, &bool_temp, &loser_total);
    or(&selection, &this_warn, &this_warn, &bool_temp);

    loser_nagents = 2;

    for(j = 0;
	(toroidal
	 ? loser_toroidalnbr_arr[i][j]
	 : loser_planarnbr_arr[i][j]) != -1;
	j++) {
      loser_nagents++;
      add(&selection, &loser_total,
	  &losers_wealth_arr[toroidal
			    ? loser_toroidalnbr_arr[i][j]
			    : loser_planarnbr_arr[i][j]]);

      printf(", %d", (toroidal
		      ? loser_toroidalnbr_arr[i][j]
		      : loser_planarnbr_arr[i][j]));

      check_warn(&selection, &bool_temp, &loser_total);
      or(&selection, &this_warn, &this_warn, &bool_temp);
    }

    printf("; nbrs (inc self) = %d\n", loser_nagents);

    set(&selection, &expr_temp, &losers_wealth_arr[i]);
    cmul(&selection, &expr_temp, (long double)(loser_nagents));
    check_warn(&selection, &bool_temp, &expr_temp);
    or(&selection, &this_warn, &this_warn, &bool_temp);

    gt(&selection, &awareness, &expr_temp, &loser_total);
    check_warn(&selection, &bool_temp, &expr_temp);
    or(&selection, &this_warn, &this_warn, &bool_temp);

    /* print a warning message if a new warning has been issued */

    not(&selection, &bool_temp, &warnings);
    and(&selection, &bool_temp, &bool_temp, &this_warn);
    qprint(&selection, &bool_temp,
	   "warning issued for loser awareness condition (after last cycle)");

    or(&selection, &warnings, &warnings, &this_warn);
				/* update warnings.
				 *
				 * IMPORTANT NOTE: Warnings after the last
				 * cycle do not impact the validity of the
				 * run. The warning messages are only
				 * printed for information.
				 */

    not(&selection, &bool_temp, &awareness);
    and(&selection, &this_valid, &this_valid, &bool_temp);
				/* update validity of this part of the cycle */

    /* print a warning message if the awareness condition is satisfied */

    not(&selection, &bool_temp, &this_valid);
    and(&selection, &bool_temp, &bool_temp, &validity);
    qprint(&selection, &bool_temp,
	   "loser awareness condition satisfied (after last cycle)");

    and(&selection, &validity, &validity, &this_valid);
				/* update validity of the run */

    kill(&selection, &loser_total);
  }

  not(&selection, &bool_temp, &warnings);
  and(&selection, &validity, &validity, &bool_temp);
  
  not(&selection, &bool_temp, &validity);
  qprint(&selection, &bool_temp, "run invalid or contained warnings");

  qprint(&selection, &validity, "run completed successfully");

  /* print wealths if required */

  if(show_wealth) {
    print_values(&selection, &winner_wealth, "Winner's wealth:");

    for(i = 0; i < NAGENTS - 1; i++) {
      printf("Loser %d's ", i);
      print_values(&selection, &losers_wealth_arr[i], "wealth:");
    }
  }

  /* free all the created objects */

  kill(&selection, &total_wealth);
  for(i = 0; i < NAGENTS - 1; i++) {
    kill(&selection, &losers_wealth_arr[i]);
  }
  kill(&selection, &expr_temp);
  kill(&selection, &ticket_price);
  kill(&selection, &losers_wealth);
  kill(&selection, &winner_wealth);
  kill(&selection, &generosity);
  kill(&selection, &jackpot);

  return 0;
}
