/*
    CharityWorld: ObserverSwarm.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

#import "ObserverSwarm.h"
#import "ModelSwarm.h"
#import "Verbosity.h"
#import "DoubleSimple.h"
#import "PNGRaster.h"
#import "Agent.h"
#import <collections.h>
#import <objectbase.h>
#import <analysis.h>
#import <gui.h>
#import <math.h>

@implementation ObserverSwarm

-createEnd {
  [super createEnd];
  termTimeSpecified = NO;
  return self;
}

-setParameterFile: (const char *)filename {
  parameterFile = filename;
  return self;
}

-setWealthPNGSave: (int)value fileStem: (const char *)stem {
  saveWealthPNG = YES;
  saveWealthPNGInterval = value;
  wealthPNGStem = stem;
  return self;
}

-setTermTime: (int)terminationTime {
  termTimeSpecified = YES;
  termTime = terminationTime;
  return self;
}

-setBlackAndWhiteRasters: (BOOL)value {
  bwrasters = value;
  return self;
}

-setScheduleGraphFileOutput: (const char *)filename {
  scheduleGraphFileOutput = YES;
  scheduleGraphFileName = filename;
  return self;
}

-buildObjects {
  int i;
  double min, max, neut, absmin, absmax;

  if([Verbosity showProgress]) printf("Building observer objects\n");
  [super buildObjects];
  displayFrequency = 100;

  // Load in the parameter file using a Swarm Archiver object

  if([Verbosity showProgress]) {
    printf("Loading model from file %s\n", parameterFile);
  }
  archiver = [LispArchiver create: self setPath: parameterFile];

  model = [archiver getWithZone: self key: "modelSwarm"];
  if(model == nil) {
    raiseEvent(InvalidOperation, "Can't find archiver file for modelSwarm");
  }
  if([Verbosity showProgress]) printf("Model loaded successfully\n");

  // Build probe displays for the modelSwarm and observerSwarm, then set
  // the control panel to "Stop" so that the user can change any parameters
  // on the GUI before running the model

  CREATE_ARCHIVED_PROBE_DISPLAY(model);
  CREATE_ARCHIVED_PROBE_DISPLAY(self);

  if([Verbosity showProgress]) {
    printf("Awaiting user input from ProcCtrl\n");
  }

  [controlPanel setStateStopped];

  // Build the model objects

  [model buildObjects];

  if([Verbosity showProgress]) {
    printf("Built model\n");
  }

  // Build the GUI components

  // Build a colour map. This has 0-63 in various shades of red with a
  // bit of blue, 64 yellow, and 65-128 in various shades of green
  // with a bit of blue, 129 cyan, 130-161 various shades of pure red,
  // 162-193 various shades of pure green, and 200-210 in various
  // colours for the floating point class display.

  colormap = [Colormap create: self];
				
  for(i = 0; i < 64; i++) {
    if(bwrasters) {
      [colormap setColor: i ToRed: 0.33 Green: 0.33 Blue: 0.33];
      [colormap setColor: i + 65 ToRed: 0.66 Green: 0.66 Blue: 0.66];
    }
    else {
      [colormap setColor: i
		ToRed: 0.33 + (double)(63 - i) / 94.5
		Green: 0.0
		Blue: 0.5];
      [colormap setColor: i + 65
		ToRed: 0.0
		Green: 0.33 + (double)i / 94.5
		Blue: 0.5];
    }
  }
  if(bwrasters) {
    [colormap setColor: 64 ToRed: 1.0 Green: 1.0 Blue: 1.0];
    [colormap setColor: 129 ToRed: 0.75 Green: 0.75 Blue: 0.75];
  }
  else {
    [colormap setColor: 64 ToRed: 1.0 Green: 1.0 Blue: 0.0];
    [colormap setColor: 129 ToRed: 0.0 Green: 1.0 Blue: 1.0];
  }
  for(i = 0; i < 32; i++) {
    if(bwrasters) {
      [colormap setColor: i + 130 ToRed: 0.0 Green: 0.0 Blue: 0.0];
      [colormap setColor: i + 162 ToRed: 0.85 Green: 0.85 Blue: 0.85];
    }
    else {
      [colormap setColor: i + 130
		ToRed: 0.33 + (double)(31 - i) / 46.5
		Green: 0.0
		Blue: 0.0];
      [colormap setColor: i + 162
		ToRed: 0.0
		Green: 0.33 + (double)i / 46.5
		Blue: 0.0];
    }
  }
  if(bwrasters) {
    [colormap setColor: 200 ToRed: 0.0 Green: 0.0 Blue: 0.0];
    [colormap setColor: 201 ToRed: 0.0 Green: 0.0 Blue: 0.0];
    [colormap setColor: 202 ToRed: 0.0 Green: 0.0 Blue: 0.0];
    [colormap setColor: 203 ToRed: 0.0 Green: 0.0 Blue: 0.0];
    [colormap setColor: 204 ToRed: 0.0 Green: 0.0 Blue: 0.0];
				// Unknown, NaN, Inf: Black

    [colormap setColor: 205 ToRed: 0.66 Green: 0.66 Blue: 0.66];
    [colormap setColor: 206 ToRed: 0.66 Green: 0.66 Blue: 0.66];
				// Denormalised: Light grey

    [colormap setColor: 207 ToRed: 1.0 Green: 1.0 Blue: 1.0];
    [colormap setColor: 208 ToRed: 1.0 Green: 1.0 Blue: 1.0];
				// Zero: White

    [colormap setColor: 209 ToRed: 0.33 Green: 0.33 Blue: 0.33];
    [colormap setColor: 210 ToRed: 0.33 Green: 0.33 Blue: 0.33];
				// Normalised: Dark grey
  }
  else {
    [colormap setColor: 200 ToRed: 0.0 Green: 0.0 Blue: 0.0];
				// Default: Black -- fpclass unknown

    [colormap setColor: 201 ToRed: 1.0 Green: 0.0 Blue: 1.0];
				// Signalling NaN: Magenta

    [colormap setColor: 202 ToRed: 1.0 Green: 1.0 Blue: 1.0];
				// Quiet NaN: White

    [colormap setColor: 203 ToRed: 0.0 Green: 0.0	Blue: 1.0];
				// Negative infinity: Blue

    [colormap setColor: 204 ToRed: 1.0 Green: 0.0 Blue: 0.0];
				// Positive infinity: Red

    [colormap setColor: 205 ToRed: 0.0 Green: 0.66 Blue: 1.0];
				// Negative denormalised: Bluey-cyan

    [colormap setColor: 206 ToRed: 1.0 Green: 0.66 Blue: 0.0];
				// Positive denormalised: Reddish-orange

    [colormap setColor: 207 ToRed: 0.0 Green: 1.0 Blue: 1.0];
				// Negative zero: Cyan

    [colormap setColor: 208 ToRed: 1.0 Green: 1.0 Blue: 0.0];
				// Positive zero: Yellow

    [colormap setColor: 209 ToRed: 0.0 Green: 0.33 Blue: 1.0];
				// Negative normalised: Very bluey cyan

    [colormap setColor: 210 ToRed: 1.0 Green: 0.33 Blue: 0.0];
				// Positive normalised: Very reddish orange
  }

  if(!hideIlogbRaster) {
    ilogbmap = [Colormap create: self];

    for(i = 0; i < 50; i++) {
      if(bwrasters) {
	[ilogbmap setColor: i ToRed: 0.8 Green: 0.8 Blue: 0.8];
				// +ve number, +ve exponent: light light grey

	[ilogbmap setColor: i + 50 ToRed: 0.5 Green: 0.5 Blue: 0.5];
				// +ve number, -ve exponent: light dark grey

	[ilogbmap setColor: i + 100 ToRed: 0.4 Green: 0.4 Blue: 0.4];
				// -ve number, -ve exponent: dark dark grey

	[ilogbmap setColor: i + 150 ToRed: 0.7 Green: 0.7 Blue: 0.7];
				// -ve number, +ve exponent: dark light grey
	
      }
      else {
	[ilogbmap setColor: i
		  ToRed: 0.33 + (double)(49 - i) / 73.5
		  Green: 0.0
		  Blue: 0.0];
				// +ve number, +ve exponent: Shades of red

	[ilogbmap setColor: i + 50
		  ToRed: 0.33 + (double)(49 - i) / 73.5
		  Green: 0.0
		  Blue: 0.33 + (double)(49 - i) / 73.5];
				// +ve number, -ve exponent: Shades of magenta

	[ilogbmap setColor: i + 100
		  ToRed: 0.0
		  Green: 0.0
		  Blue: 0.33 + (double)(49 - i) / 73.5];
				// -ve number, -ve exponent: Shades of blue

	[ilogbmap setColor: i + 150
		  ToRed: 0.0
		  Green: 0.33 + (double)(49 - i) / 73.5
		  Blue: 0.33 + (double)(49 - i) / 73.5];
				// -ve number, -ve exponent: Shades of cyan
      }
    }
    if(bwrasters) {
      [ilogbmap setColor: 200 ToRed: 1.0 Green: 1.0 Blue: 1.0];
				// Zero: White

      [ilogbmap setColor: 201 ToRed: 0.9 Green: 0.9 Blue: 0.9];
				// Positive minimum exponent: light light light

      [ilogbmap setColor: 202 ToRed: 0.3 Green: 0.3 Blue: 0.3];
				// Negative minimum exponent: dark dark dark

      [ilogbmap setColor: 203 ToRed: 0.0 Green: 0.0 Blue: 0.0];
      [ilogbmap setColor: 204 ToRed: 0.0 Green: 0.0 Blue: 0.0];
				// Not a number, unknown: Black

    }
    else {
      [ilogbmap setColor: 200 ToRed: 1.0 Green: 1.0 Blue: 0.0];
				// Zero: Yellow

      [ilogbmap setColor: 201 ToRed: 1.0 Green: 0.5 Blue: 0.0];
				// Positive minimum exponent: Reddish-orange

      [ilogbmap setColor: 202 ToRed: 0.0 Green: 0.5 Blue: 1.0];
				// Negative minimum exponent: Bluey-cyan

      [ilogbmap setColor: 203 ToRed: 1.0 Green: 1.0 Blue: 1.0];
				// Not a number: White

      [ilogbmap setColor: 204 ToRed: 0.0 Green: 0.0 Blue: 0.0];
				// Unknown: Black
    }
  }

  if([Verbosity showProgress]) {
    printf("Built colormap(s)\n");
  }

  // Build a raster to show absolute wealth of agents in various shades.

  wealthRaster = [ZoomRaster createBegin: self];
  SET_WINDOW_GEOMETRY_RECORD_NAME(wealthRaster);
  wealthRaster = [wealthRaster createEnd];
  [wealthRaster enableDestroyNotification: self
                notificationMethod: M(quitWealthRaster:)];
  [wealthRaster setColormap: colormap];
  [wealthRaster setZoomFactor: 9];
  [wealthRaster setWidth: [model getXSize] Height: [model getYSize]];
  [wealthRaster setWindowTitle: "Wealth World"];
  [wealthRaster pack];

  wealthDisplay = [Object2dDisplay create: self
                                   setDisplayWidget: wealthRaster
                                   setDiscrete2dToDisplay: [model getWorld]
                                   setDisplayMessage: M(drawWealthOn:)];
  [wealthDisplay setObjectCollection: [model getAgents]];

  donationTraceDisplay = [Object2dDisplay create: self
					  setDisplayWidget: wealthRaster
					  setDiscrete2dToDisplay:
					    [model getWorld]
					  setDisplayMessage:
					    M(drawTracedDonationsOn:)];
  [donationTraceDisplay setObjectCollection: [model getAgents]];

  [wealthRaster setButton: ButtonLeft
                Client: wealthDisplay
                Message: M(makeProbeAtX:Y:)];

  if(saveWealthPNG) {
    wealthPNG = [PNGRaster create: self setRaster: wealthRaster];
  }

  if([Verbosity showProgress]) {
    printf("Built wealth raster\n");
  }

  // Build a raster to show relative (local) wealth of agents in various shades

  localWealthRaster = [ZoomRaster createBegin: self];
  SET_WINDOW_GEOMETRY_RECORD_NAME(localWealthRaster);
  localWealthRaster = [localWealthRaster createEnd];
  [localWealthRaster enableDestroyNotification: self
                notificationMethod: M(quitLocalWealthRaster:)];
  [localWealthRaster setColormap: colormap];
  [localWealthRaster setZoomFactor: 9];
  [localWealthRaster setWidth: [model getXSize] Height: [model getYSize]];
  [localWealthRaster setWindowTitle: "Local Wealth World"];
  [localWealthRaster pack];

  localWealthDisplay = [Object2dDisplay create: self
					setDisplayWidget: localWealthRaster
					setDiscrete2dToDisplay:
					  [model getWorld]
					setDisplayMessage:
					  M(drawLocalWealthOn:)];
  [localWealthDisplay setObjectCollection: [model getAgents]];

  [localWealthRaster setButton: ButtonLeft
                Client: localWealthDisplay
                Message: M(makeProbeAtX:Y:)];

  if([Verbosity showProgress]) {
    printf("Built local wealth raster\n");
  }

  // Build a raster to show the floating point class of the wealth
#ifndef __CYGWIN__
  fpclassRaster = [ZoomRaster createBegin: self];
  SET_WINDOW_GEOMETRY_RECORD_NAME(fpclassRaster);
  fpclassRaster = [fpclassRaster createEnd];
  [fpclassRaster enableDestroyNotification: self
		 notificationMethod: M(quitFPClassRaster:)];
  [fpclassRaster setColormap: colormap];
  [fpclassRaster setZoomFactor: 9];
  [fpclassRaster setWidth: [model getXSize] Height: [model getYSize]];
  [fpclassRaster setWindowTitle: "Wealth FP Class"];
  [fpclassRaster pack];

  fpclassDisplay = [Object2dDisplay create: self
				    setDisplayWidget: fpclassRaster
				    setDiscrete2dToDisplay: [model getWorld]
				    setDisplayMessage:
				      M(drawFloatingPointClassOn:)];
  if([Verbosity showProgress]) {
    printf("Built fpclass raster\n");
  }
#endif

  if(!hideIlogbRaster) {
    // Build a raster to show the log2 of wealth

    ilogbRaster = [ZoomRaster createBegin: self];
    SET_WINDOW_GEOMETRY_RECORD_NAME(ilogbRaster);
    ilogbRaster = [ilogbRaster createEnd];
    [ilogbRaster enableDestroyNotification: self
		 notificationMethod: M(quitIlogbRaster:)];
    [ilogbRaster setColormap: ilogbmap];
    [ilogbRaster setZoomFactor: 9];
    [ilogbRaster setWidth: [model getXSize] Height: [model getYSize]];
    [ilogbRaster setWindowTitle: "ilogb(wealth)"];
    [ilogbRaster pack];

    ilogbDisplay = [Object2dDisplay create: self
				    setDisplayWidget: ilogbRaster
				    setDiscrete2dToDisplay: [model getWorld]
				    setDisplayMessage: M (drawLog2WealthOn:)];

    if([Verbosity showProgress]) {
      printf("Built ilogb raster\n");
    }
  }

  // Build a raster to show membership of the beggar and donor schedules

  scheduleRaster = [ZoomRaster createBegin: self];
  SET_WINDOW_GEOMETRY_RECORD_NAME(scheduleRaster);
  scheduleRaster = [scheduleRaster createEnd];
  [scheduleRaster enableDestroyNotification: self
		  notificationMethod: M(quitScheduleRaster:)];
  [scheduleRaster setColormap: colormap];
  [scheduleRaster setZoomFactor: 9];
  [scheduleRaster setWidth: [model getXSize] Height: [model getYSize]];
  [scheduleRaster setWindowTitle: "Beggar/Donor Schedules"];
  [scheduleRaster pack];

  scheduleDisplay = [Object2dDisplay create: self
				     setDisplayWidget: scheduleRaster
				     setDiscrete2dToDisplay: [model getWorld]
				     setDisplayMessage: M(drawScheduleOn:)];

  if([Verbosity showProgress]) {
    printf("Built schedule raster\n");
  }

  // Build a time-series graph showing the wealth of the agents

  wealthGraph = [EZGraph create: self
                         setTitle: "Wealth of agents vs. time"
                         setAxisLabelsX: "Time" Y: "Wealth"
                         setWindowGeometryRecordName: "wealthGraph"];
  [wealthGraph enableDestroyNotification: self
               notificationMethod: M(quitWealthGraph:)];
  if(!hideWealthMax) {
    [wealthGraph createMaxSequence: "max"
		 withFeedFrom: [model getAgents]
		 andSelector: M(getWealthForDisplayOnly)];
  }
  if(!hideWealthMean) {
    [wealthGraph createAverageSequence: "pop mean"
		 withFeedFrom: [model getAgents]
		 andSelector: M(getWealthForDisplayOnly)];
  }
  if(!hideBeggarWealthMean) {
    [wealthGraph createAverageSequence: "beggar mean"
		 withFeedFrom: [[model getBeggars] getList]
		 andSelector: M(getWealthForDisplayOnly)];
  }
  if(!hideDonorWealthMean) {
    [wealthGraph createAverageSequence: "donor mean"
		 withFeedFrom: [[model getDonors] getList]
		 andSelector: M(getWealthForDisplayOnly)];
  }
  if(!hideWealthMin) {
    [wealthGraph createMinSequence: "min"
		 withFeedFrom: [model getAgents]
		 andSelector: M(getWealthForDisplayOnly)];
  }

  if([Verbosity showProgress]) {
    printf("Built wealth graph\n");
  }

  // Build a graph showing the size of the beggar and donor schedules against
  // time.

  beggarGraph = [EZGraph create: self
                         setTitle: "Size of schedule lists vs. time"
                         setAxisLabelsX: "Time" Y: "Size"
                         setWindowGeometryRecordName: "beggarGraph"];
  [beggarGraph enableDestroyNotification: self
               notificationMethod: M(quitBeggarGraph:)];
  if(scheduleGraphFileOutput) {
    [beggarGraph setFileOutput: YES];
    [beggarGraph setFileName: scheduleGraphFileName];
  }
  if(!hideScheduleBeggars) {
    [beggarGraph createSequence: "beggars"
		 withFeedFrom: [model getBeggars]
		 andSelector: M(getCount)];
  }
  if(!hideScheduleDonors) {
    [beggarGraph createSequence: "donors"
		 withFeedFrom: [model getDonors]
		 andSelector: M(getCount)];
  }

  if([Verbosity showProgress]) {
    printf("Built schedule list graph\n");
  }

  // Build a distribution graph of beggar wealth

  min = [model getMinWealthForDisplayOnly];
  max = [model getMaxWealthForDisplayOnly];
  neut = [model getInitialWealthForDisplayOnly];

  absmin = (fabs(neut - min) > fabs(max - neut))
    ? neut - 1.5 * fabs(max - neut) : neut - 1.5 * fabs(neut - min);
  absmax = (fabs(neut - min) > fabs(max - neut))
    ? neut + 1.5 * fabs(max - neut) : neut + 1.5 * fabs(neut - min);

  if(absmin == absmax) {
    absmin = min;
    absmax = max;
  }

  beggarDist = [EZBin createBegin: self];
  [beggarDist setCollection: [[model getBeggars] getList]];
  [beggarDist setProbedSelector: M(getWealthForDisplayOnly)];
  [beggarDist setTitle: "Beggar wealth distribution"];
  [beggarDist setAxisLabelsX: "Wealth" Y: "Number"];
  [beggarDist setBinCount: 12];
  [beggarDist setMonoColorBars: YES];
  [beggarDist setLowerBound: absmin];
  [beggarDist setUpperBound: absmax];
  beggarDist = [beggarDist createEnd];
  [beggarDist enableDestroyNotification: self
              notificationMethod: M(quitBeggarDist:)];

  // Build a distribution graph of donor wealth

  donorDist = [EZBin createBegin: self];
  [donorDist setCollection: [[model getDonors] getList]];
  [donorDist setProbedSelector: M(getWealthForDisplayOnly)];
  [donorDist setTitle: "Donor wealth distribution"];
  [donorDist setAxisLabelsX: "Wealth" Y: "Number"];
  [donorDist setBinCount: 10];
  [donorDist setMonoColorBars: YES];
  [donorDist setLowerBound: absmin];
  [donorDist setUpperBound: absmax]; 
  donorDist = [donorDist createEnd];
  [donorDist enableDestroyNotification: self
              notificationMethod: M(quitDonorDist:)];

  if([Verbosity showProgress]) {
    printf("Built wealth distribution graphs\n");
  }

  // Build a graph showing the number of items in various Swarm Zones each
  // time step (useful for tracing memory leaks).

  zoneGraph = [EZGraph create: self
		       setTitle: "Size of various zones vs. time"
		       setAxisLabelsX: "Time" Y: "Size"
		       setWindowGeometryRecordName: "zoneGraph"];
  [zoneGraph enableDestroyNotification: self
	     notificationMethod: M(quitZoneGraph:)];
  [zoneGraph createSequence: "scratch"
	     withFeedFrom: self
	     andSelector: M(countScratchZone)];
  [zoneGraph createSequence: "observer"
	     withFeedFrom: self
	     andSelector: M(countObserverZone)];
  [zoneGraph createSequence: "model"
	     withFeedFrom: self
	     andSelector: M(countModelZone)];
  [zoneGraph createSequence: "double"
	     withFeedFrom: self
	     andSelector: M(countDoubleSimples)];

  if([Verbosity showProgress]) {
    printf("Built zone graph\n");
  }
  
  return self;
}

/*

updateDisplay

This method is called from the observer schedule to update the displays of 
all the various graphs and rasters.

*/

-updateDisplay {
  static int counter = 0;

  if(wealthRaster != nil) {
    [wealthDisplay display];
    [donationTraceDisplay display];
    [wealthRaster drawSelf];
    [[model getAgents] forEach: M(clearDonatedNeighbours)];
  }
  if(localWealthRaster != nil) {
    [localWealthDisplay display];
    [localWealthRaster drawSelf];
  }
#ifndef __CYGWIN__
  if(fpclassRaster != nil) {
    [fpclassDisplay display];
    [fpclassRaster drawSelf];
  }
#endif
  if(ilogbRaster != nil) {
    [ilogbDisplay display];
    [ilogbRaster drawSelf];
  }
  if(scheduleRaster != nil) {
    [scheduleDisplay display];
    [scheduleRaster drawSelf];
  }
  if(wealthGraph != nil) {
    [wealthGraph step];
  }
  if(beggarGraph != nil) {
    [beggarGraph update];
    [beggarGraph outputGraph];
    if(scheduleGraphFileOutput) {
      [beggarGraph outputToFile];
    }
  }
  if(beggarDist != nil) {
    [beggarDist reset];
    [beggarDist update];
    [beggarDist outputGraph];
  }
  if(donorDist != nil) {
    [donorDist reset];
    [donorDist update];
    [donorDist outputGraph];
  }
  if(zoneGraph != nil) {
    [zoneGraph step];
  }
  [probeDisplayManager update];
  [actionCache doTkEvents];
  if(wealthRaster != nil) {
    if(saveWealthPNG) {
      if(counter % saveWealthPNGInterval == 0) {
	[wealthPNG saveToFileStem: wealthPNGStem];
      }
    }
  }
  counter++;
  return self;
}

-buildActions {
  if([Verbosity showProgress]) printf("Building observer schedule\n");
  [super buildActions];
  [model buildActions];

  displaySchedule = [Schedule create: self
			      setRepeatInterval: displayFrequency];
  [displaySchedule at: 0 createActionTo: self message: M(updateDisplay)];
  
  if(termTimeSpecified) {
    outerSchedule = [Schedule createBegin: self];
    outerSchedule = [outerSchedule createEnd];
    [outerSchedule at: termTime
		   createActionTo: controlPanel
		   message: M(setStateStopped)];
  }	    

  return self;
}

-activateIn: swarmContext {
  [super activateIn: swarmContext];
  [displaySchedule activateIn: self];
  if(outerSchedule!=nil)[outerSchedule activateIn: self];
  [model activateIn: self];
  return [self getActivity];
}

-(double)countScratchZone {
  return (double)[self countZone: scratchZone];
}

-(double)countObserverZone {
  return (double)[self countZone: self];
}

-(double)countModelZone {
  return (double)[self countZone: model];
}

-(double)countDoubleSimples {
  return (double)[DoubleSimple getCounter];
}

-(unsigned long long)countZone: (id <Zone>)z {
  id <List> pop = [z getPopulation];
  id <ListIndex> ix = [pop listBegin: scratchZone];
  unsigned long long c = 0;
  id member;

  for(member = [ix next]; [ix getLoc] == Member; member = [ix next]) {
    if([member isKindOfClassNamed: "Zone_c"]) {
      c += [self countZone: member];
    }
    else {
      c++;
    }
  }
  [ix drop];
  return c;
}

/*

These quit... methods are required to allow the program to cope with the
various widgets created to be destroyed by the user during the simulation.
This then stops them from being updated in the updateDisplay method.

*/

-quitWealthRaster: caller {
  [wealthRaster drop];
  wealthRaster = nil;
  return self;
}

-quitLocalWealthRaster: caller {
  [localWealthRaster drop];
  localWealthRaster = nil;
  return self;
}

#ifndef __CYGWIN__
-quitFPClassRaster: caller {
  [fpclassRaster drop];
  fpclassRaster = nil;
  return self;
}
#endif

-quitIlogbRaster: caller {
  [ilogbRaster drop];
  ilogbRaster = nil;
  return self;
}

-quitScheduleRaster: caller {
  [scheduleRaster drop];
  scheduleRaster = nil;
  return self;
}

-quitWealthGraph: caller {
  [wealthGraph drop];
  wealthGraph = nil;
  return self;
}

-quitBeggarGraph: caller {
  [beggarGraph drop];
  beggarGraph = nil;
  return self;
}

-quitBeggarDist: caller {
  [beggarDist drop];
  beggarDist = nil;
  return self;
}

-quitDonorDist: caller {
  [donorDist drop];
  donorDist = nil;
  return self;
}

-quitZoneGraph: caller {
  [zoneGraph drop];
  zoneGraph = nil;
  return self;
}

@end

